@@ -1,5 +1,7 @@
 from __future__ import annotations
 
+import concurrent.futures
+
 import functools
 
 import dataclasses
@@ -16,7 +18,6 @@ from pathlib import Path
 from typing import Generic, TypeVar, Dict
 
 import faiss
-import gitignore_parser
 import numpy as np
 import tiktoken
 from tqdm import tqdm
@@ -89,6 +90,12 @@ class OpenAIRequestScheduler:
         self.batch_tokens = 0
         self.batch_size_limit = 2048
         self.batch_token_limit = 400_000
+        self.max_concurrent_requests = 8
+        self.executor = concurrent.futures.ThreadPoolExecutor(
+            max_workers=self.max_concurrent_requests,
+            thread_name_prefix="OpenAIRequestScheduler"
+        )
+        self.futures = []
 
     @cached_property
     def encoding(self) -> tiktoken.Encoding:
@@ -98,25 +105,40 @@ class OpenAIRequestScheduler:
         request_tokens = len(self.encoding.encode(req.content, disallowed_special=()))
         results = []
         if request_tokens > 8192:
+            STATS.increment("codebased.embeddings.skipped.too_long")
             return results
         if len(self.batch) >= self.batch_size_limit or self.batch_tokens + request_tokens > self.batch_token_limit:
-            results = self.flush()
-            self.batch.clear()
+            self.futures.append(self.executor.submit(self._process_batch, self.batch))
+            self.batch = []
             self.batch_tokens = 0
         self.batch.append(req)
         self.batch_tokens += request_tokens
+
+        futures = []
+        for future in self.futures:
+            if future.done():
+                # This may raise an exception that should propagate.
+                results.extend(future.result())
+            else:
+                futures.append(future)
+        self.futures = futures
+
         return results
 
     def flush(self) -> T.Iterable[Embedding]:
-        if not self.batch:
-            return []
-        results = create_openai_embeddings_sync_batched(
-            self.oai_client,
-            self.batch,
-            self.embedding_config
-        )
-        self.batch = []
-        self.batch_tokens = 0
+        if self.batch:
+            self.futures.append(self.executor.submit(self._process_batch, self.batch))
+            self.batch = []
+            self.batch_tokens = 0
+
+        results = []
+        for future in concurrent.futures.as_completed(self.futures):
+            results.extend(future.result())
+        self.futures.clear()
+        return results
+
+    def _process_batch(self, batch: T.List[EmbeddingRequest]) -> T.Iterable[Embedding]:
+        results = create_openai_embeddings_sync_batched(self.oai_client, batch, self.embedding_config)
         return results
 
 
@@ -252,14 +274,6 @@ class Dependencies:
         migrations.migrate()
         return db
 
-    @cached_property
-    def gitignore(self) -> T.Callable[[Path], bool]:
-        gitignore_path = self.config.root / '.gitignore'
-        try:
-            return gitignore_parser.parse_gitignore(gitignore_path, base_dir=self.config.root)
-        except FileNotFoundError:
-            return lambda _: False
-
     def ignore_checker(self, path: Path) -> bool:
         original_path = path
         while path.is_relative_to(self.config.root):
@@ -558,6 +572,7 @@ def index_paths(
                 for obj_id, obj in objects_by_id.items():
                     rendered = render_object(obj, in_lines=in_lines, file=False)
                     if not rendered:
+                        STATS.increment("codebased.embeddings.skipped.empty")
                         continue
                     request = EmbeddingRequest(
                         object_id=obj_id,
